package ap.proof;

import ap.terfor.{Formula, TermOrder, Conjunction, Quantifier, NegatedConjunctions}
import ap.proof.goal.Goal
import ap.terfor.arithconj.{ArithConj, ReduceWithAC}
import ap.util.{Debug, Seqs}

object ConstraintSimplifier {
  
  val AC = Debug.AC_CONSTRAINT_SIMPLIFIER  
  
  val NO_SIMPLIFIER : ConstraintSimplifier = new ConstraintSimplifier {
    def apply(f : Formula, order : TermOrder) : Formula = f
  }
  
  val ALL_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(true, false)

  val ALL_SIMPLIFIER_TRACE : ConstraintSimplifier =
    new SimpleSimplifier(true, true)

  val QUANTIFICATION_SIMPLIFIER : ConstraintSimplifier =
    new SimpleSimplifier(false, false)

}

abstract class ConstraintSimplifier {
  /**
   * Simplify the formula <code>f</code>
   */
  def apply(f : Formula, order : TermOrder) : Formula
}

class SimpleSimplifier(groundSimplifications : Boolean, output : Boolean)
      extends ConstraintSimplifier {
  
  private lazy val prover =
    new ExhaustiveProver(ConstraintSimplifier.NO_SIMPLIFIER)

  private def print(str : => String) : unit = (if (output) Predef.print(str))
  private def println(str : => String) : unit = (if (output) Predef.println(str))
  
  private val EMPTY : Set[Quantifier]  = Set()
  private val ALL : Set[Quantifier]    = Set(Quantifier.ALL)
  private val EX : Set[Quantifier]     = Set(Quantifier.EX)
  private val ALL_EX : Set[Quantifier] = Set(Quantifier.EX, Quantifier.ALL)
    
  def apply(f : Formula, order : TermOrder) : Formula = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ConstraintSimplifier.AC, order isSortingOf f)
    ////////////////////////////////////////////////////////////////////////////

    collectQuantifiers(f) match {
      case ALL => simplify(f, order)
      case EX => negSimplify(f, order)
      case EMPTY if (groundSimplifications) => f match {
        case f : Conjunction if (f.size > 1) => negSimplify(f, order)
        case _ => f
      }
      case EMPTY => f
    }
  }

  private def simplify(f : Formula, order : TermOrder) : Formula = {
    println("Simplify: " + f + " (positive)")
    val res = prover(Goal(f, order)).closingConstraint
    println("          " + res)
    res
  }

  private def negSimplify(f : Formula, order : TermOrder) : Formula = {
    println("Simplify: " + f + " (negative)")
    val goal = Goal(Conjunction.negate(f, order), order)
    val res = Conjunction.negate(prover(goal).closingConstraint, order)    
    println("          " + res)
    res
  }
  
  private def collectQuantifiers(f : Formula) : Set[Quantifier] =
    Conjunction.collectQuantifiers(f, (conj) =>
           Set() ++ conj.quans.drop(1) ++
           // we only look for proper divisibility (not divisibility by one ...)
           (if (((conj.arithConj.positiveEqs.elements ++
                  conj.arithConj.negativeEqs.elements).next()(0) _1).isOne)
              Set(conj.quans(0))
            else
              Set()))
  
}
