package ap.proof.goal;

import ap.basetypes.IdealInt
import ap.terfor.{Formula, Conjunction, Quantifier, TermOrder}
import ap.terfor.equations.EquationConj
import ap.terfor.linearcombination.LinearCombination
import ap.terfor.preds.Atom
import ap.terfor.arithconj.ReduceWithAC
import ap.util.Debug

object AddFactsTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    (f.isTrue || f.isLiteral ||
     (f.isNegatedConjunction && f.negatedConjs(0).negatedConjs.isEmpty)) && 
    f.quans.isEmpty

}

class AddFactsTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(AddFactsTask.AC, !formula.isFalse)
  //////////////////////////////////////////////////////////////////////////////

  val priority : Int = -20000

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.isTrue) {
      // then the goal can be closed immediately
      ptf.updateGoal(Conjunction.FALSE, goal)
    } else if (formula.isLiteral) {
      addFacts(Conjunction.negate(formula, goal.order), goal, ptf)
    } else {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(AddFactsTask.AC, formula.isNegatedConjunction)
      //////////////////////////////////////////////////////////////////////////

      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(AddFactsTask.AC,
                      disj.quans.isEmpty && disj.negatedConjs.isEmpty)
      //////////////////////////////////////////////////////////////////////////
            
      // the literals of the negated conjunction can be added as facts to
      // the goal
      addFacts(disj, goal, ptf)
    }

  private def addFacts(newFacts : Conjunction,
                       goal : Goal,
                       ptf : ProofTreeFactory) : ProofTree = {
    val updatedFacts = Conjunction.conj(Array(goal.facts, newFacts), goal.order)
    val newTasks = unifyAtoms(goal.facts.predConj.positiveLits,
                              newFacts.predConj.negativeLits,
                              goal) ++
                   unifyAtoms(goal.facts.predConj.negativeLits,
                              newFacts.predConj.positiveLits,
                              goal) ++
                   unifyAtoms(newFacts.predConj.negativeLits,
                              newFacts.predConj.positiveLits,
                              goal)
    ptf.updateGoal(updatedFacts, newTasks, goal)
  }

  private def unifyAtoms(atoms1 : Seq[Atom], atoms2 : Seq[Atom],
                         goal : Goal) : Seq[PrioritisedTask] = {
    val reducer = ReduceWithAC(goal.facts.arithConj, goal.order)
    // TODO: make this more efficient
    for (a1 <- atoms1; a2 <- atoms2;
         t <- FormulaTask(reducer(unify(a1, a2, goal.order)), goal.age)) yield t
  }
  
  private def unify(a1 : Atom, a2 : Atom, order : TermOrder) : Conjunction =
    if (a1.pred == a2.pred) {
      val eqs =
        EquationConj(for ((arg1, arg2) <- (a1.elements zip a2.elements))
                     yield LinearCombination.sum(Array((IdealInt.ONE, arg1),
                                                       (IdealInt.MINUS_ONE, arg2)),
                                                 order),
                     order)
      Conjunction.conj(eqs, order)
    } else {
      Conjunction.FALSE
    }
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new AddFactsTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    AddFactsTask.isCoveredFormula(f)

  protected val name : String = "AddFacts"

}
