package ap.proof.goal;

import ap.terfor.{Formula, Conjunction, Quantifier, ConstantTerm}
import ap.terfor.arithconj.ArithConj
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.substitutions.VariableSubst
import ap.terfor.preds.PredConj
import ap.util.{Debug, PlainRange}

object BetaFormulaTask {
  
  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  def isCoveredFormula(f : Conjunction) : Boolean =
    f.size > 1 && f.quans.isEmpty

}


class BetaFormulaTask(_formula : Conjunction, _age : Int)
      extends FormulaTask(_formula, _age) {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(BetaFormulaTask.AC,
                   !formula.isTrue && !formula.isFalse &&
                   !formula.isLiteral && !formula.isNegatedConjunction)
  //////////////////////////////////////////////////////////////////////////////
      
  val priority : Int = -50 + age

  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree =
    if (formula.arithConj.isTrue && formula.predConj.isTrue) {
      // split to handle the complex conjuncts
      applyBeta(goal, ptf)
    } else {
      // cut over one of the literals
      applyCut(goal, ptf)
    }

  private def applyBeta(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, formula.negatedConjs.size > 1)
    ////////////////////////////////////////////////////////////////////////////
      
    val subtrees = for (negConj <- formula.negatedConjs)
                   yield ptf.updateGoal(FormulaTask(negConj.negate, goal.age), goal)
    ptf.and(subtrees, goal.order)
  }

  private def applyCut(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val order = goal.order    
    val (cutLiteral, rem) = selectCutLiteral(formula)
    
    val firstTree = {
      // assume that the cut-literal is true
//      val updatedFacts = Conjunction.conj(Array(goal.facts, cutLiteral), order)
      ptf.updateGoal(//updatedFacts,    // normal tableau-splitting for the time being
                     FormulaTask(rem, goal.age),
                     goal)
    }
    
    val secondTree = {
      // assume that the cut-literal is false
      ptf.updateGoal(FormulaTask(cutLiteral, goal.age), goal)
    }
    
    ptf.and(Array(secondTree, firstTree), order)    
  }
  
  /**
   * Given a conjunction of formulas, select one of the formulas
   * as literal to cut over. The result is the pair of the selected formula and
   * the remaining <code>Conjunction</code>
   */
  private def selectCutLiteral(conj : Conjunction) : (Conjunction, Conjunction) =
    if (!conj.arithConj.isTrue) {
      val (sel, rem) = selectCutLiteral(conj.arithConj)
      (Conjunction.conj(sel, conj.order), conj.updateArithConj(rem, conj.order))      
    } else {
      //////////////////////////////////////////////////////////////////////////
      Debug.assertInt(BetaFormulaTask.AC, !conj.predConj.isTrue)
      //////////////////////////////////////////////////////////////////////////
      val (sel, rem) = selectCutLiteral(conj.predConj)
      (Conjunction.conj(sel, conj.order), conj.updatePredConj(rem, conj.order))            
    }

  private def selectCutLiteral(ac : ArithConj) : (ArithConj, ArithConj) =
    if (!ac.positiveEqs.isEmpty) {
      val (sel, rem) = selectCutLiteral(ac.positiveEqs)
      (ArithConj.conj(sel, ac.order), ac.updatePositiveEqs(rem, ac.order))      
    } else if (!ac.negativeEqs.isEmpty) {
      val (sel, rem) = selectCutLiteral(ac.negativeEqs)
      (ArithConj.conj(sel, ac.order), ac.updateNegativeEqs(rem, ac.order))
    } else {
      val (sel, rem) = selectCutLiteral(ac.inEqs)
      (ArithConj.conj(sel, ac.order), ac.updateInEqs(rem, ac.order))
    }

  private def selectCutLiteral(eqs : EquationConj) : (EquationConj, EquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (EquationConj(eqs(eqs.size - 1), eqs.order),
     EquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }

  private def selectCutLiteral(eqs : NegEquationConj)
                                      : (NegEquationConj, NegEquationConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !eqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (NegEquationConj(eqs(eqs.size - 1), eqs.order),
     NegEquationConj(eqs.slice(0, eqs.size - 1), eqs.order))
  }
  
  private def selectCutLiteral(inEqs : InEqConj) : (InEqConj, InEqConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !inEqs.isEmpty)
    ////////////////////////////////////////////////////////////////////////////
    (InEqConj(inEqs(inEqs.size - 1), inEqs.order),
     InEqConj(inEqs.slice(0, inEqs.size - 1), inEqs.order))
  }
  
  private def selectCutLiteral(conj : PredConj) : (PredConj, PredConj) = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(BetaFormulaTask.AC, !conj.isTrue)
    ////////////////////////////////////////////////////////////////////////////
    val posLits = conj.positiveLits
    val negLits = conj.negativeLits
    if (!posLits.isEmpty)
      (PredConj(List(posLits(posLits.size - 1)), List(), conj.order),
       PredConj(posLits.slice(0, posLits.size - 1), negLits, conj.order))
    else
      (PredConj(List(), List(negLits(negLits.size - 1)),conj.order),
       PredConj(List(), negLits.slice(0, negLits.size - 1), conj.order))
  }

  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask =
    new BetaFormulaTask(f, age)

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean =
    BetaFormulaTask.isCoveredFormula(f)

  protected val name : String = "BetaFor"

}
