package ap.proof.goal;

import ap.terfor.{Conjunction, Quantifier}
import ap.terfor.arithconj.ReduceWithAC
import ap.util.Debug

object FormulaTask {

  private val AC = Debug.AC_COMPLEX_FORMULAS_TASK
  
  def apply(formula : Conjunction, age : Int) : Iterable[FormulaTask] =
    if (formula.isFalse) {
      List()
    } else if (formula.isTrue || formula.isLiteral) {
      List(new AddFactsTask(formula, age))
    } else if (formula.isNegatedConjunction) {
      val disj = formula.negatedConjs(0)
      //////////////////////////////////////////////////////////////////////////
      // possibly existing quantifiers should have been pulled out and
      // should not occur at this point
      Debug.assertInt(FormulaTask.AC, disj.quans.isEmpty)
      //////////////////////////////////////////////////////////////////////////

      (for (f <- disj.negatedConjs; g <- FormulaTask(f, age)) yield g) ++
      (if (disj.arithConj.isTrue && disj.predConj.isTrue)
         List()
       else
         List(new AddFactsTask(Conjunction.negate(Conjunction.conj(Array(disj.arithConj,
                                                                         disj.predConj),
                                                                   formula.order),
                                                  formula.order),
                               age)))
    } else if (formula.quans.isEmpty) {
      List(new BetaFormulaTask(formula, age))
    } else formula.quans.last match {
      case Quantifier.ALL => List(new AllQuantifierTask(formula, age))
      case Quantifier.EX =>
        if (formula.isDivisibility)
          List(new DivisibilityTask(formula, age))
        else
          List(new ExQuantifierTask(formula, age))
    }
  
}


/**
 * The representation of formulas in a proof goal that are more complex than
 * simple facts. Such formulas are considered to have positive
 * polarity, i.e., as conjunctions in the succedent of a goal.
 * This class is both responsible for storing such formulas and for eventually
 * processing the formulas, e.g. by splitting up the formulas/proof goal.
 */
abstract class FormulaTask(val formula : Conjunction, val age : Int)
         extends PrioritisedTask {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(FormulaTask.AC, isCoveredFormula(formula))
  //////////////////////////////////////////////////////////////////////////////

  /**
   * Return <code>true</code> if <code>f</code> is a formula that can be handled
   * by this task
   */
  protected def isCoveredFormula(f : Conjunction) : Boolean
  
  /**
   * Perform the actual task (whatever needs to be done with <code>formula</code>)
   */
  protected def doTask(goal : Goal, ptf : ProofTreeFactory) : ProofTree
  
  /**
   * Create a new <code>FormulaTask</code> by updating the value of
   * <code>formula</code>
   */
  protected def updateFormula(f : Conjunction) : FormulaTask
                                                                  
  def apply(goal : Goal, ptf : ProofTreeFactory) : ProofTree = {
    val reducedFormula = ReduceWithAC(goal.facts.arithConj, goal.order)(formula)
    if (reducedFormula == formula) {
      doTask(goal, ptf)
    } else if (isCoveredFormula(reducedFormula)) {
      this.updateFormula(reducedFormula).doTask(goal, ptf)
    } else {
      ptf.updateGoal(FormulaTask(reducedFormula, age), goal)
    }
  }
  
  protected val name : String
   
  override def toString : String = name + "(" + formula + ")"
   
}
