package ap.proof.goal;

import ap.util.Debug
import ap.basetypes.{LeftistHeap, HeapCollector}

object TaskManager {
  
  private def AC = Debug.AC_GOAL
  
  private implicit def orderTask(thisTask : PrioritisedTask) : Ordered[PrioritisedTask] =
    new Ordered[PrioritisedTask] {
      def compare(thatTask : PrioritisedTask) : Int =
        (thisTask.priority compare thatTask.priority)
    }
  
  protected[goal] type TaskHeap = LeftistHeap[PrioritisedTask, TaskInfoCollector]
    
  //////////////////////////////////////////////////////////////////////////////
  
  val EMPTY : TaskManager =
    new TaskManager (LeftistHeap.EMPTY_HEAP(TaskInfoCollector.EMPTY),
                     EagerTaskManager.INITIAL)
  
}

/**
 * An immutable class (priority queue) for handling a set of tasks in a proof
 * goal. Currently, this is implemented using a sorted set, but it would be
 * better to use a real immutable queue (leftist heap?).
 *
 * TODO: So far, no real subsumption checks are performed
 */
class TaskManager private (// the regular tasks that have a priority
                           prioTasks : TaskManager.TaskHeap,
                           // Preprocessing tasks that can sneak in before
                           // regular tasks.
                           eagerTasks : EagerTaskManager) {

  def +(t : PrioritisedTask) = new TaskManager (prioTasks + t, eagerTasks)

  def ++ (elems: Iterable[PrioritisedTask]): TaskManager =
    this ++ elems.elements

  def ++ (elems: Iterator[PrioritisedTask]): TaskManager =
    (this /: elems) ((manager, elem) => manager + elem)

  def enqueue(elems: PrioritisedTask*): TaskManager = (this ++ elems)

  /**
   * Remove the first task from the queue.
   */
  def removeFirst : TaskManager = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(TaskManager.AC, !isEmpty)
    ////////////////////////////////////////////////////////////////////////////

    nextEagerTask match {
      case None =>
        new TaskManager (prioTasks.deleteMin, eagerTasks afterTask prioTasks.findMin)
      case Some(task) =>
        new TaskManager (prioTasks, eagerTasks afterTask task)
    }
  }
  
  private def nextEagerTask : Option[EagerTask] =
    eagerTasks recommend prioTasks.findMinOption
  
  /** Returns the element with the smallest priority value in the queue,
   *  or throws an error if there is no element contained in the queue.
   *
   *  @return   the element with the smallest priority.
   */
  def max: Task = nextEagerTask getOrElse prioTasks.findMin

  /**
   * Compute information about the prioritised tasks (eager tasks are not
   * considered at this point)
   */
  def taskInfos : TaskInfoCollector = prioTasks.collector
  
  //////////////////////////////////////////////////////////////////////////////

  def isEmpty : Boolean = prioTasks.isEmpty && nextEagerTask.isEmpty
    
  def prioritisedTasks : Iterable[PrioritisedTask] = prioTasks
  
  //////////////////////////////////////////////////////////////////////////////
  
  override def toString : String = {
    val strings =
      for (t <- nextEagerTask.elements ++ prioTasks.elements) yield t.toString

    "[" + (if (strings.hasNext)
             strings.reduceLeft((s1 : String, s2 : String) => s1 + ", " + s2)
           else
             "") + "]"
  }
  
}
