package ap.terfor.arithconj;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj, ReduceWithEqs, ReduceWithNegEqs}
import ap.terfor.inequalities.{InEqConj, ReduceWithInEqs}
import ap.terfor.preds.{Atom, PredConj}
import ap.terfor.substitutions.VariableShiftSubst
import ap.util.{Debug, LazyMappedSet, UnionSet, Seqs, FilterIt}

object ReduceWithAC {
  
  private val AC = Debug.AC_PROPAGATION

  def apply(ac : ArithConj, order : TermOrder) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, ac isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    new ReduceWithAC(ReduceWithEqs(ac.positiveEqs, order),
                     ReduceWithNegEqs(ac.negativeEqs, order),
                     ReduceWithInEqs(ac.inEqs, order),
                     order)
  }

  //////////////////////////////////////////////////////////////////////////////
  // Some of the "static" methods of the <code>ReduceWithAC</code>-class
  // These are methods juggling with different reducer-objects
      
  /**
   * Reduce a conjunction of arithmetic stuff and return the reduced conjunction,
   * together with a new <code>ReduceWithAC</code> object to which the reduced
   * conjunction was added.
   */
  private def reduceAC(ac : ArithConj,
                       initialReducer : ReduceWithAC) : (ArithConj, ReduceWithAC) = {

    // positive equations always come first
    val newPosEqs = initialReducer.reduce(ac.positiveEqs, ac.inEqs.equalityInfs)
    var reducer = initialReducer addEquations newPosEqs
      
    // then, reduce inequalities, assuming the (unreduced) negated equations
    val newInEqs = (reducer addEquations ac.negativeEqs) reduce ac.inEqs
    reducer = reducer addInEqs newInEqs
      
    // reduce negated equations, assuming the reduced inequalities
    val newNegEqs = reducer reduce ac.negativeEqs
    reducer = reducer addEquations newNegEqs
      
    val newAC = ArithConj(newPosEqs, newNegEqs, newInEqs, reducer.order)
    if (newInEqs.equalityInfs.isEmpty && newNegEqs == ac.negativeEqs)
      (newAC, reducer)
    else
      // if the new inequalities still imply equations, we have to reduce once
      // more. note, that we again start with the reducer <code>initialReducer</code> 
      reduceAC(newAC, initialReducer)
  }

  /**
   * Recursively reduce an arbitrary conjunction
   */
  private def reduceConj(conj : Conjunction,
                         initialReducer : ReduceWithAC) : Conjunction = try {
    val (newArithConj, reducer) =
      ReduceWithAC.reduceAC(conj.arithConj,
                            initialReducer.passQuantifiers(conj.quans.size))

    val newPredConj = reducer reduce conj.predConj
                            
    val newNegConjs = NegatedConjunctions(for (c <- conj.negatedConjs.elements)
                                          yield (reduceConj(c, reducer)),
                                          reducer.order)
     
    val res = Conjunction(conj.quans, newArithConj, newPredConj, newNegConjs, reducer.order)
     
    if (newArithConj == res.arithConj) {
      res
    } else {
      // it might be necessary to repeat reduction, because new facts became
      // available
      reduceConj(res, initialReducer)
    }
  } catch {
    case FALSE_EXCEPTION => Conjunction.FALSE
  }

}

class ReduceWithAC private (positiveEqs : ReduceWithEqs,
                            negativeEqs : ReduceWithNegEqs,
                            inEqs : ReduceWithInEqs,
                            private val order : TermOrder) {

  /**
   * Create a <code>ReduceWithEqs</code> that can be used underneath
   * <code>num</code> binders. The conversion of de Brujin-variables is done on
   * the fly, which should give a good performance when the resulting
   * <code>ReduceWithEqs</code> is not applied too often (TODO: caching)
   */
  private def passQuantifiers(num : Int) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, num >= 0)
    ////////////////////////////////////////////////////////////////////////////
    if (num == 0)
      this
    else
      new ReduceWithAC(positiveEqs passQuantifiers num,
                       negativeEqs passQuantifiers num,
                       inEqs passQuantifiers num,
                       order)
  }

  private def addEquations(eqs : EquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (eqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs addEquations eqs.toMap,
                       negativeEqs,
                       inEqs,
                       order)
  }
  
  private def addEquations(eqs : NegEquationConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (eqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs,
                       negativeEqs addEquations eqs.toSet,
                       inEqs,
                       order)
  }
  
  private def addInEqs(furtherInEqs : InEqConj) : ReduceWithAC = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, furtherInEqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    if (furtherInEqs.isEmpty)
      this
    else
      new ReduceWithAC(positiveEqs,
                       negativeEqs,
                       inEqs addInEqs furtherInEqs,
                       order)    
  }
  
  //////////////////////////////////////////////////////////////////////////////
  // Methods for reducing different kinds of formulas. If any of the
  // methods detects that a resulting formula is false,
  // <code>FALSE_EXCEPTION</code> is thrown (this is done to simplify the
  // handling of <code>Conjunction</code>s)
  // TODO: optimise for cases where no reduction is possible (detect this early,
  // create no new objects)
  // TODO: define the following methods in a nicer way, polymorphic?
  
  private def reduce(_eqs : EquationConj*) : EquationConj = {
    val eqs = EquationConj.conj(_eqs, order)
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = inEqs(negativeEqs(positiveEqs(eqs)))
    if (redEqs.isFalse) throw FALSE_EXCEPTION
    redEqs
  }

  private def reduce(eqs : NegEquationConj) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, eqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redEqs = inEqs(negativeEqs(positiveEqs(eqs)))
    if (redEqs.isFalse) throw FALSE_EXCEPTION
    redEqs
  }

  private def reduce(ies : InEqConj) : InEqConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, ies isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redInEqs = inEqs(negativeEqs(positiveEqs(ies)))
    if (redInEqs.isFalse) throw FALSE_EXCEPTION
    redInEqs
  }

  private def reduce(a : Atom) : Atom = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, a isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    Atom(a.pred, for (lc <- a.elements) yield positiveEqs(lc), order)
  }

  private def reduce(conj : PredConj) : PredConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithAC.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    val redConj = PredConj(for (a <- conj.positiveLits.elements) yield reduce(a),
                           for (a <- conj.negativeLits.elements) yield reduce(a),
                           order)
    if (redConj.isFalse) throw FALSE_EXCEPTION
    redConj
  }

  private def reduce(conj : ArithConj) : ArithConj =
    try { ReduceWithAC.reduceAC(conj, this) _1 }
    catch { case FALSE_EXCEPTION => ArithConj.FALSE }

  //////////////////////////////////////////////////////////////////////////////
    
  def apply(conj : ArithConj) : ArithConj =  {
    val res = reduce(conj)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithAC.AC, reduce(res) == res)
    ////////////////////////////////////////////////////////////////////////////
    res
  }
  
  def apply(conj : Conjunction) : Conjunction = {
    val res = ReduceWithAC.reduceConj(conj, this)
    ////////////////////////////////////////////////////////////////////////////
    // we demand that the reducer is a projection (repeated application does not
    // change the result anymore)
    Debug.assertPost(ReduceWithAC.AC, ReduceWithAC.reduceConj(res, this) == res)
    ////////////////////////////////////////////////////////////////////////////
    res
  }
  
}

object FALSE_EXCEPTION extends Exception

