package ap.terfor.inequalities;

import scala.collection.mutable.ArrayBuffer

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.basetypes.IdealInt
import ap.terfor.substitutions.VariableShiftSubst
import ap.util.{Debug, Logic, Seqs, FilterIt}

object ReduceWithInEqs {
  
  private val AC = Debug.AC_PROPAGATION

  def apply(inEqs : InEqConj, order : TermOrder) : ReduceWithInEqs = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(AC, inEqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    new ReduceWithInEqs(inEqs.findLowerBound _, order)
  }
  
}

/**
 * Reduce certain terms or formulas drawing information from inequalities that
 * are assumed as facts
 */
class ReduceWithInEqs private (lowerBound : (LinearCombination) => Option[IdealInt],
                               order : TermOrder) {

  def addInEqs(furtherInEqs : InEqConj) : ReduceWithInEqs = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithInEqs.AC, furtherInEqs isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    new ReduceWithInEqs((lc:LinearCombination) => (
                          // we compute the maximum of all known lower bounds
                          (lowerBound(lc), furtherInEqs findLowerBound lc) match {
                            case (Some(c), Some(d)) => Some(c max d)
                            case (x@Some(_), _) => x
                            case (_, x@Some(_)) => x
                            case _ => None
                          }),
                        order)    
  }
  
  /**
   * Create a <code>ReduceWithEqs</code> that can be used underneath
   * <code>num</code> binders. The conversion of de Brujin-variables is done on
   * the fly, which should give a good performance when the resulting
   * <code>ReduceWithEqs</code> is not applied too often (TODO: caching)
   */
  def passQuantifiers(num : Int) : ReduceWithInEqs = {
    val downShifter = VariableShiftSubst.downShifter[LinearCombination](num, order)
    new ReduceWithInEqs((lc:LinearCombination) => (if (downShifter isDefinedAt lc)
                                                     lowerBound(downShifter(lc))
                                                   else
                                                     None),
                         order)
  }

  
  def apply(conj : EquationConj) : EquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithInEqs.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////

    // the only possible inference is that the conjunction of equations is
    // unsatisfiable
    if (Logic.exists(for (lc <- conj.elements) yield isNonZero(lc)))
      EquationConj.FALSE
    else
      conj
  }

  /**
   * Determine whether we can derive from the known inequalities that the value
   * of <code>lc</code> is not zero
   */
  private def isNonZero(lc : LinearCombination) : Boolean =
    // yes, it has to be twice isPositive!
    isPositive(lowerBound(lc)) || isPositive(lowerBound(-lc))
    
  private def isPositive : Option[IdealInt] => Boolean = {
    case Some(d) => d.signum > 0
    case _ => false
  }
  
  
  /**
   * Reduce a conjunction of negated equations by removing all equations from
   * which we know that they hold anyway
   */
  def apply(conj : NegEquationConj) : NegEquationConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithInEqs.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    conj.updateEqs(conj filter ((lc:LinearCombination) => !isNonZero(lc)),
                   order)
  }

  /**
   * Reduce a conjunction of inequalities. This means that subsumed inequalities
   * are removed, contradictions are detected, and possibly further equations
   * are inferred.
   */
  def apply(conj : InEqConj) : InEqConj = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(ReduceWithInEqs.AC, conj isSortedBy order)
    ////////////////////////////////////////////////////////////////////////////
    
    try {
      conj.updateGeqZero(for (lc <- conj.elements; newLC <- refine(lc))
                         yield newLC,
                         order)
    } catch {
      case CONTRADICTION_EXCEPTION => InEqConj.FALSE
    }
  }

  /**
   * Refine a single geq-zero-inequality using the assumed inequalities. If a
   * contradiction is detected, the exception
   * <code>CONTRADICTION_EXCEPTION</code> is raised. 
   */
  private def refine(lc : LinearCombination) : Iterator[LinearCombination] =
    lowerBound(-lc) match {
    case Some(d) if (d.signum > 0) => 
      // contradiction
      throw CONTRADICTION_EXCEPTION
    case x => // we also need to check lower bounds
              (lowerBound(lc), x) match {
              case (Some(d), _) if (d.signum >= 0) =>
                // the inequality is subsumed by a known inequality,
                // can be removed
                Iterator.empty
              case (_, Some(IdealInt.ZERO)) =>
                // we can infer an equation from an inequality by inserting an
                // upper bound as well
                Array(lc, -lc).elements
              case _ =>
                // we have to keep the inequality
                Iterator.single(lc)              
              }
    }

}

private object CONTRADICTION_EXCEPTION extends Exception
