package ap.terfor.substitutions;

import ap.terfor.linearcombination.LinearCombination
import ap.terfor.equations.{EquationConj, NegEquationConj}
import ap.terfor.inequalities.InEqConj
import ap.terfor.arithconj.ArithConj
import ap.terfor.preds.{Atom, PredConj}
import ap.util.Debug

object Substitution {

  private val AC = Debug.AC_SUBSTITUTIONS
  
}

/**
 * A substitution is a mapping from <code>TerFor</code> to <code>TerFor</code>
 * that replaces variables and constants with arbitrary other terms. It is
 * required that a substitution is only applied to terms/formulas that are
 * sorted according to <code>order</code>. There are two more concrete
 * sub-traits: <code>SimpleSubstitution</code>, which performs a simple
 * replacement of constants or variables, and <code>PseudoDivSubstitution</code>,
 * which can make use of pseudo-division in order to replace expressions
 * <code>n * c</code>.
 */
trait Substitution extends ((TerFor) => TerFor) with Sorted[Substitution] {

  /**
   * The term order that is used for the resulting terms or formulas. We require
   * that a substitution is only applied to terms/formulas that are already
   * sorted according to this order
   */
  protected[substitutions] def order : TermOrder
  
  /**
   * Substitution that is to be used underneath <code>num</code> quantifiers.
   * Because we use De Bruijn indexes, passing quantifiers shifts the variables
   * in a substitution
   */
  protected[substitutions] def passQuantifiers(num : Int) : Substitution
  
  /**
   * For performance reasons, a <code>Substitution</code> can tell whether
   * it is possible at all its application will change a term or a formula. This
   * can be found out by looking at the free variables in an expression, for
   * instance. In this case, this method has to return <code>false</code>.
   */
  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean

  protected def idOrElse[A <: TerFor](t : A, otherwise : => A) : A = {
    ////////////////////////////////////////////////////////////////////////////
    Debug.assertPre(Substitution.AC, order isSortingOf t)
    ////////////////////////////////////////////////////////////////////////////     
     
    if (isIdentityOn(t)) t else otherwise     
  }

  //////////////////////////////////////////////////////////////////////////////
  
  /**
   * Compare the order of this <code>Substitution</code> with a given order. We
   * use equality here, because the behaviour would be quite confusing with
   * the relation <code>isSubOrderOf</code> (remember that the substitution has
   * to cope with arbitrary terms/formulas that are sorted by the order) 
   */
  def isSortedBy(otherOrder : TermOrder) : Boolean = (order == otherOrder)

  //////////////////////////////////////////////////////////////////////////////
  
  def apply(t : TerFor) : TerFor = {
    t match {
    case t : Term => apply(t)
    case f : Formula => apply(f)
    }
  }

  //////////////////////////////////////////////////////////////////////////////
  // The following methods are implemented in the sub-traits
    
  def apply(t : Term) : Term

  def apply(lc : LinearCombination) : LinearCombination

  def apply(conj : EquationConj) : EquationConj
  
  def apply(negConj : NegEquationConj) : NegEquationConj

  def apply(negConj : InEqConj) : InEqConj

  //////////////////////////////////////////////////////////////////////////////
  
  def apply(f : Formula) : Formula = {
    f match {
    case f : EquationConj => apply(f)
    case f : NegEquationConj => apply(f)
    case f : InEqConj => apply(f)
    case f : ArithConj => apply(f)
    case f : Conjunction => apply(f)
    case f : NegatedConjunctions => apply(f)
    }
  }
  
  def apply(conj : ArithConj) : ArithConj =
    idOrElse(conj,
             ArithConj(apply(conj.positiveEqs),
                       apply(conj.negativeEqs),
                       apply(conj.inEqs),
                       order))

  def apply(conj : Conjunction) : Conjunction =
    idOrElse(conj,
             {
               val pastSubst = this.passQuantifiers(conj.quans.size)
               Conjunction(conj.quans,
                           pastSubst(conj.arithConj),
                           pastSubst(conj.predConj),
                           pastSubst(conj.negatedConjs),
                           order)
             })

  def apply(conjs : NegatedConjunctions) : NegatedConjunctions =
    idOrElse(conjs,
             NegatedConjunctions(for (conj <- conjs) yield apply(conj), order))

  def apply(a : Atom) : Atom =
    idOrElse(a,
             Atom(a.pred, for (lc <- a.elements) yield apply(lc), order))

  def apply(conj : PredConj) : PredConj =
    idOrElse(conj,
             PredConj(for (a <- conj.positiveLits.elements) yield apply(a),
                      for (a <- conj.negativeLits.elements) yield apply(a),
                      order))
}
